@@ -46,18 +46,7 @@ STDOUT_LINE = '\nStdout:\n%s'
 STDERR_LINE = '\nStderr:\n%s'
 
 
-def xml_safe_unicode(base, encoding='utf-8'):
-    """Return a unicode string containing only valid XML characters.
-
-    encoding - if base is a byte string it is first decoded to unicode
-        using this encoding.
-    """
-    if isinstance(base, six.binary_type):
-        base = base.decode(encoding)
-    return INVALID_XML_1_0_UNICODE_RE.sub('', base)
-
-
-def to_unicode(data):
+def _to_unicode(data):
     """Returns unicode in Python2 and str in Python3"""
     if six.PY3:
         return six.text_type(data)
@@ -68,8 +57,17 @@ def to_unicode(data):
         return repr(data).decode('utf8', 'replace')
 
 
-def safe_unicode(data, encoding=None):
-    return xml_safe_unicode(to_unicode(data), encoding)
+def safe_unicode(data, encoding='utf8'):
+    """Return a unicode string containing only valid XML characters.
+
+    encoding - if data is a byte string it is first decoded to unicode
+        using this encoding.
+    """
+    data = _to_unicode(data)
+    if isinstance(data, six.binary_type):
+        # e.g. IronPython, see #182
+        data = data.decode(encoding)
+    return INVALID_XML_1_0_UNICODE_RE.sub('', data)
 
 
 def testcase_name(test_method):
@@ -431,7 +429,7 @@ class _XMLTestResult(_TextTestResult):
         tests_by_testcase = {}
 
         for tests in (self.successes, self.failures, self.errors,
-                      self.skipped):
+                      self.skipped, self.expectedFailures, self.unexpectedSuccesses):
             for test_info in tests:
                 if isinstance(test_info, tuple):
                     # This is a skipped, error or a failure test case
@@ -462,9 +460,12 @@ class _XMLTestResult(_TextTestResult):
         """
         testsuite = xml_document.createElement('testsuite')
         parentElement.appendChild(testsuite)
+        module_name = suite_name.rpartition('.')[0]
+        file_name = module_name.replace('.', '/') + '.py'
 
         testsuite.setAttribute('name', suite_name)
         testsuite.setAttribute('tests', str(len(tests)))
+        testsuite.setAttribute('file', file_name)
 
         testsuite.setAttribute(
             'time', '%.3f' % sum(map(lambda e: e.elapsed_time, tests))
@@ -488,28 +489,6 @@ class _XMLTestResult(_TextTestResult):
         for test in tests:
             _XMLTestResult._report_testcase(test, testsuite, xml_document)
 
-        systemout = xml_document.createElement('system-out')
-        testsuite.appendChild(systemout)
-
-        stdout = StringIO()
-        for test in tests:
-            # Merge the stdout from the tests in a class
-            if test.stdout is not None:
-                stdout.write(test.stdout)
-        _XMLTestResult._createCDATAsections(
-            xml_document, systemout, stdout.getvalue())
-
-        systemerr = xml_document.createElement('system-err')
-        testsuite.appendChild(systemerr)
-
-        stderr = StringIO()
-        for test in tests:
-            # Merge the stderr from the tests in a class
-            if test.stderr is not None:
-                stderr.write(test.stderr)
-        _XMLTestResult._createCDATAsections(
-            xml_document, systemerr, stderr.getvalue())
-
         return testsuite
 
     _report_testsuite = staticmethod(_report_testsuite)
@@ -572,6 +551,18 @@ class _XMLTestResult(_TextTestResult):
                 failure.setAttribute('type', 'skip')
                 failure.setAttribute('message', test_result.test_exception_message)
 
+        if test_result.stdout:
+            systemout = xml_document.createElement('system-out')
+            testcase.appendChild(systemout)
+            _XMLTestResult._createCDATAsections(
+                xml_document, systemout, test_result.stdout)
+
+        if test_result.stderr:
+            systemout = xml_document.createElement('system-err')
+            testcase.appendChild(systemout)
+            _XMLTestResult._createCDATAsections(
+                xml_document, systemout, test_result.stderr)
+
     _report_testcase = staticmethod(_report_testcase)
 
     def generate_reports(self, test_runner):
@@ -607,12 +598,12 @@ class _XMLTestResult(_TextTestResult):
             testsuite = _XMLTestResult._report_testsuite(
                 suite_name, tests, doc, parentElement, self.properties
             )
-            xml_content = doc.toprettyxml(
-                indent='\t',
-                encoding=test_runner.encoding
-            )
 
             if outputHandledAsString:
+                xml_content = doc.toprettyxml(
+                    indent='\t',
+                    encoding=test_runner.encoding
+                )
                 filename = path.join(
                     test_runner.output,
                     'TEST-%s.xml' % suite_name)
@@ -623,6 +614,10 @@ class _XMLTestResult(_TextTestResult):
 
         if not outputHandledAsString:
             # Assume that test_runner.output is a stream
+            xml_content = doc.toprettyxml(
+                indent='\t',
+                encoding=test_runner.encoding
+            )
             test_runner.output.write(xml_content)
 
     def _exc_info_to_string(self, err, test):
